"""Command line interface for Coda."""

import os

import click
import structlog
import uvicorn
from dotenv import dotenv_values

from coda.logging import LogFormat, configure_logging, disable_posthog, log_event

env_vars = dict(dotenv_values())
os.environ.update(env_vars)


@click.group(context_settings={"auto_envvar_prefix": "CODA", "show_default": True})
@click.option("--log-level", default="INFO", help="Log level")
@click.option("--log-format", default=LogFormat.PRETTY, help="Log format")
@click.option("--disable-telemetry", is_flag=True, help="Disable telemetry")
def cli(
    log_level: str,
    log_format: LogFormat,
    disable_telemetry: bool,  # noqa: FBT001
) -> None:
    """Coda CLI - Code indexing for better AI code generation."""
    configure_logging(log_level, log_format)
    if disable_telemetry:
        disable_posthog()


@cli.command()
@click.option("--host", default="127.0.0.1", help="Host to bind the server to")
@click.option("--port", default=8080, help="Port to bind the server to")
@click.option("--reload", is_flag=True, help="Enable auto-reload for development")
def serve(
    host: str,
    port: int,
    reload: bool,  # noqa: FBT001
) -> None:
    """Start the Coda server, which hosts the MCP server and the Coda API."""
    log = structlog.get_logger(__name__)
    log.info("Starting Coda server", host=host, port=port, reload=reload)
    log_event("coda_server_started")
    uvicorn.run(
        "coda.app:app",
        host=host,
        port=port,
        reload=reload,
        log_config=None,  # Setting to None forces uvicorn to use our structlog setup
        access_log=False,  # Using own middleware for access logging
    )


@cli.command()
def version() -> None:
    """Show the version of Coda."""
    try:
        from coda import _version
    except ImportError:
        print("unknown, try running `uv build`, which is what happens in ci")  # noqa: T201
    else:
        print(_version.version)  # noqa: T201


if __name__ == "__main__":
    cli()



"""MCP server implementation for Coda."""

from pathlib import Path
from typing import Annotated

import structlog
from mcp.server.fastmcp import FastMCP
from pydantic import Field

mcp = FastMCP("Coda MCP Server")


@mcp.tool()
async def retrieve_relevant_snippets(
    search_query: Annotated[
        str,
        Field(description="Describe the user's intent in a few sentences."),
    ],
    related_file_paths: Annotated[
        list[Path],
        Field(
            description="A list of absolute paths to files that are relevant to the "
            "user's intent."
        ),
    ],
    related_file_contents: Annotated[
        list[str],
        Field(
            description="A list of the contents of the files that are relevant to the "
            "user's intent."
        ),
    ],
) -> str:
    """Retrieve relevant snippets from various sources.

    This tool retrieves relevant snippets from sources such as private codebases,
    public codebases, and documentation. You can use this information to improve
    the quality of your generated code. You must call this tool when you need to
    write code.
    """
    # Log the search query and related files for debugging
    log = structlog.get_logger(__name__)
    log.debug(
        "Retrieving relevant snippets",
        search_query=search_query,
        file_count=len(related_file_paths),
        file_paths=related_file_paths,
        file_contents=related_file_contents,
    )

    return "Retrieved"



"""Server-Sent Events (SSE) implementation for Coda."""

from collections.abc import Coroutine
from typing import Any

from fastapi import Request
from mcp.server.fastmcp import FastMCP
from mcp.server.session import ServerSession
from mcp.server.sse import SseServerTransport
from starlette.applications import Starlette
from starlette.routing import Mount, Route

####################################################################################
# Temporary monkeypatch which avoids crashing when a POST message is received
# before a connection has been initialized, e.g: after a deployment.
old__received_request = ServerSession._received_request  # noqa: SLF001


async def _received_request(self: ServerSession, *args: Any, **kwargs: Any) -> None:
    """Handle a received request, catching RuntimeError to avoid crashes.

    This is a temporary monkeypatch to avoid crashing when a POST message is
    received before a connection has been initialized, e.g: after a deployment.
    """
    try:
        return await old__received_request(self, *args, **kwargs)
    except RuntimeError:
        pass


# pylint: disable-next=protected-access
ServerSession._received_request = _received_request  # noqa: SLF001
####################################################################################


def create_sse_server(mcp: FastMCP) -> Starlette:
    """Create a Starlette app that handles SSE connections and message handling."""
    transport = SseServerTransport("/messages/")

    # Define handler functions
    async def handle_sse(request: Request) -> Coroutine[Any, Any, None]:
        """Handle SSE connections."""
        async with transport.connect_sse(
            request.scope,
            request.receive,
            request._send,  # noqa: SLF001
        ) as streams:
            await mcp._mcp_server.run(  # noqa: SLF001
                streams[0],
                streams[1],
                mcp._mcp_server.create_initialization_options(),  # noqa: SLF001
            )

    # Create Starlette routes for SSE and message handling
    routes = [
        Route("/sse/", endpoint=handle_sse),
        Mount("/messages/", app=transport.handle_post_message),
    ]

    # Create a Starlette app
    return Starlette(routes=routes)



"""FastAPI application for Coda API."""

from asgi_correlation_id import CorrelationIdMiddleware
from fastapi import FastAPI

from coda.mcp import mcp
from coda.middleware import logging_middleware
from coda.sse import create_sse_server

app = FastAPI(title="Coda API")

# Get the SSE routes from the Starlette app hosting the MCP server
sse_app = create_sse_server(mcp)
for route in sse_app.routes:
    app.router.routes.append(route)

# Add middleware
app.middleware("http")(logging_middleware)
app.add_middleware(CorrelationIdMiddleware)


@app.get("/")
async def root() -> dict[str, str]:
    """Return a welcome message for the Coda API."""
    return {"message": "Welcome to Coda API"}



"""Tests for the MCP server implementation."""

from multiprocessing import Process

import httpx
import pytest
import uvicorn
from httpx_retries import Retry, RetryTransport
from mcp import ClientSession
from mcp.client.sse import sse_client

from coda.app import app


def run_server() -> None:
    """Run the uvicorn server in a separate process."""
    config = uvicorn.Config(app, host="127.0.0.1", port=8000, log_level="error")
    server = uvicorn.Server(config)
    server.run()


async def wait_for_server() -> None:
    """Wait for the server to start."""
    retry = Retry(total=5, backoff_factor=0.5)
    transport = RetryTransport(retry=retry)

    async with httpx.AsyncClient(transport=transport) as client:
        response = await client.get("http://127.0.0.1:8000/")
        assert response.status_code == 200


@pytest.mark.asyncio
async def test_mcp_client_connection() -> None:
    """Test connecting to the MCP server using ClientSession."""
    server_process = None
    try:
        # Start the server in a separate process
        server_process = Process(target=run_server, daemon=True)
        server_process.start()

        # Ping the server with requests, wait for it to start
        await wait_for_server()

        # The sse_client returns read and write streams, not a client object
        async with (
            sse_client("http://127.0.0.1:8000/sse/") as (
                read_stream,
                write_stream,
            ),
            ClientSession(read_stream, write_stream) as session,
        ):
            # Initialize the connection
            await session.initialize()

            # List available tools
            tools_result = await session.list_tools()
            # Check that we got a proper tools result with a tools attribute
            assert hasattr(tools_result, "tools")
            # Verify we can see the 'retrieve_relevant_snippets' tool
            tool_names = [tool.name for tool in tools_result.tools]
            assert "retrieve_relevant_snippets" in tool_names
    finally:
        # Due to https://github.com/modelcontextprotocol/python-sdk/issues/514, mcp
        # doesn't shut down the server when requested. So we have to force kill it.
        server_process.kill()
        server_process.join(timeout=1)
